#include <sys/types.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/errno.h>
#include <netinet/in.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>

#define	QLEN		   5	/* maximum connection queue length	*/
#define	BUFSIZE		16384

extern int	errno;

void	reaper(int);
int	proxyechod(int client_fd);
int	errexit(const char *format, ...);
int	passiveTCP(const char *service, int qlen);
int     connectTCP(const char *host, const char *service);
int     buf_replace(char* buf, int buf_size, char* string, char* replacement);
void    print_buf(char* buf, int upto);

char* real_proxy = (char*)0;
char* real_port = (char*)0;

char      buf[BUFSIZE];
char      rep_buf[BUFSIZE*2];
int       rep_buf_pos = 0;
char      temp[BUFSIZE];
int       pos = 0;


int main(int argc, char *argv[]){
  char	*service = "http";	/* service name or port number	*/
  struct	sockaddr_in fsin;	/* the address of a client	*/
  int	alen;			/* length of client's address	*/
  int	msock;			/* master server socket		*/
  int	ssock;			/* slave server socket		*/
  /*
  char  test1[] = {'a', 'a', 'b', 'a', 'b', 'c', 'b', 'c', 'a', 'b'};
  char  test2[] = {'c', 'b', 'a', 'b'};
  char  test3[] = {'b', 'c', 'd', 'e', 'a', 'b', 'd'};
  char  test4[] = {'a', 'b', 'c', 'd', 'e'};
  int i=0;

  print_buf(test1, 10);
  print_buf(test2, 4);
  print_buf(test3, 7);
  print_buf(test4, 5);

  printf("\n\n");

  i = buf_replace(test1, 10, "abc", "H");
  print_buf(rep_buf, i);
  i = buf_replace(test2, 4, "abc", "H");
  print_buf(rep_buf, i);
  i = buf_replace(test3, 7, "abc", "H");
  print_buf(rep_buf, i);
  i = buf_replace(test4, 5, "abc", "H");
  print_buf(rep_buf, i);

  printf("\n\n");

  return 0;
  */

  switch (argc) {
  case 4:
    real_port = argv[3];
  case 3:
    service = argv[2];
  case 2:
    real_proxy = argv[1];
    break;
  default:
    errexit("usage: proxy <realProxy> [server port] [realProxy port]\n");
  }

  msock = passiveTCP(service, QLEN);

  (void) signal(SIGCHLD, reaper);

  while (1) {
    alen = sizeof(fsin);
    ssock = accept(msock, (struct sockaddr *)&fsin, &alen);
    if (ssock < 0) {
      if (errno == EINTR)
	continue;
      errexit("accept: %s\n", strerror(errno));
    }
    switch (fork()) {
    case 0:		/* child */
      (void) close(msock);
      exit(proxyechod(ssock));
    default:	/* parent */
      (void) close(ssock);
      break;
    case -1:
      errexit("fork: %s\n", strerror(errno));
    }
  }
}

void sock_printf(int fd, const char* format, ...){
  va_list	args;
  char temp[BUFSIZE];
  int size;
  
  va_start(args, format);
  size = vsprintf(temp, format, args);
  va_end(args);

  printf(temp);
  fflush(stdout);
  write(fd, temp, size);
}


int flush_buf_replace(){
  int oldpos = pos;
  // CASE 2 - flush temp to rep_buf
  strcpy(&rep_buf[rep_buf_pos], temp);
  rep_buf_pos += pos;
  pos = 0;
  return oldpos;
}


int buf_replace(char* buf, int buf_size, char* string, char* replacement){
  int buf_pos = 0;
  int slength = strlen(string);
  int rlength = strlen(replacement);
  int tlength = 0;
  int progress;

  while(buf_pos < buf_size){
    // CASE 1 - add char to temp
    assert(pos < slength);

    if(string[pos] == buf[buf_pos]){
      temp[pos] = string[pos];
      buf_pos++;
      pos++;
      if(pos == slength){
	strcpy(&rep_buf[rep_buf_pos], replacement);
	rep_buf_pos += rlength;
	pos = 0;
      }
      continue;
    }

    // CASE 2 - flush temp to rep_buf
    assert(string[pos] != buf[buf_pos]);

    temp[pos] = buf[buf_pos];
    buf_pos++;
    pos++;
    strcpy(&rep_buf[rep_buf_pos], temp);
    rep_buf_pos += pos;
    pos = 0;
  }

  progress = rep_buf_pos;
  rep_buf_pos = 0;
  return progress;
}


void print_buf(char* buf, int upto){
  int i;
  for(i=0; i<upto; i++)
    if(iscntrl(buf[i]) || isprint(buf[i]))
      fputc(buf[i], stdout);
  //    else
  //  printf("{0x%x}", buf[i]);

  //printf(" ");
  //printf("\n\n");
}


void transfer(int proxy_fd, int client_fd){
  fd_set rfds;
  int nfds = getdtablesize();
  struct timeval base = {30, 0};	/* timeout after 30 seconds */
  struct timeval cur = base;
  int retval;
  int bytes_read = 0;
  int bytes_replaced = 0;
  int echo_fd = fileno(stdout);

  while(1){
    FD_ZERO(&rfds);
    FD_SET(proxy_fd, &rfds);
    FD_SET(client_fd, &rfds);

    retval = select(nfds, &rfds, NULL, NULL, &cur);
    cur = base;
    if(!retval)
      break;
    
    if(FD_ISSET(proxy_fd, &rfds)){
      bytes_read = read(proxy_fd, buf, BUFSIZE);

      if(bytes_read <= 0)
	break;

      bytes_replaced = buf_replace(buf, bytes_read,
				   "Proxy-Authenticate: NTLM\r\n",
				   "");
      

      //write(echo_fd, rep_buf, bytes_replaced);
      
      if(write(client_fd, rep_buf, bytes_replaced) <= 0)
	break;
    }
      
    if(FD_ISSET(client_fd, &rfds)){
      bytes_read = read(client_fd, buf, BUFSIZE);

      if(bytes_read <= 0)
	break;

      /*
      bytes_replaced = buf_replace(buf, bytes_read,
				   "HTTP/1.1\r\n",
				   "HTTP/1.0\r\nProxy-Connection: Keep-Alive\r\n");

      memcpy(buf, rep_buf, bytes_replaced);

      bytes_replaced = buf_replace(buf, bytes_replaced,
				   "Connection: keep-alive\r\n",
				   "");
      */

      //write(echo_fd, buf, bytes_read);

      if(write(proxy_fd, buf, bytes_read) <= 0)
	break;
    } 
  }

  //printf("session complete\n");
  //fflush(stdout);

  close(proxy_fd);
  close(client_fd);
}


int proxyechod(int client_fd){
  int proxy_fd = connectTCP(real_proxy, real_port);

  transfer(proxy_fd, client_fd);

  return 0;
}


/*------------------------------------------------------------------------
 * reaper - clean up zombie children
 *------------------------------------------------------------------------
 */
/*ARGSUSED*/
void reaper(int sig){
  int	status;

  while (wait3(&status, WNOHANG, (struct rusage *)0) >= 0)
    /* empty */;
}
