#include <sys/types.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/errno.h>
#include <netinet/in.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#define	QLEN		   5	/* maximum connection queue length	*/
#define	BUFSIZE		4096

extern int	errno;

void	reaper(int);
int	proxyechod(int client_fd);
int	errexit(const char *format, ...);
int	passiveTCP(const char *service, int qlen);
int     connectTCP(const char *host, const char *service);

char* real_proxy = (char*)0;
char* real_port = "80";

int main(int argc, char *argv[]){
  char	*service = "http";	/* service name or port number	*/
  struct	sockaddr_in fsin;	/* the address of a client	*/
  int	alen;			/* length of client's address	*/
  int	msock;			/* master server socket		*/
  int	ssock;			/* slave server socket		*/

  switch (argc) {
  case 4:
    service = argv[3];
  case 3:
    real_port = argv[2];
  case 2:
    real_proxy = argv[1];
    break;
  default:
    errexit("usage: proxy <realProxy> [server port] [realProxy port]\n");
  }

  msock = passiveTCP(service, QLEN);

  (void) signal(SIGCHLD, reaper);

  while (1) {
    alen = sizeof(fsin);
    ssock = accept(msock, (struct sockaddr *)&fsin, &alen);
    if (ssock < 0) {
      if (errno == EINTR)
	continue;
      errexit("accept: %s\n", strerror(errno));
    }
    switch (fork()) {
    case 0:		/* child */
      (void) close(msock);
      exit(proxyechod(ssock));
    default:	/* parent */
      (void) close(ssock);
      break;
    case -1:
      errexit("fork: %s\n", strerror(errno));
    }
  }
}

int ready(int fd, int time_usec){
  fd_set rfds;
  struct timeval wait = {0, time_usec};
  
  FD_ZERO(&rfds);
  FD_SET(fd, &rfds);
  return select(fd+1, &rfds, NULL, NULL, &wait);
}

int ready_now(int fd){
  return ready(fd, 0);
}

int has_eof(char* buf, int size){
  int i;
  for(i=0; i<size; i++){
    if(buf[i] == (char)(-1))
      return 1;
  }

  return 0;
}


int found_eof = 0;


int pipe_async(int from_fd, int to_fd){
  char buf[BUFSIZE];
  int bytes_read = 0;
  int did_something = 0;

  while(ready_now(from_fd)){
    bytes_read = read(from_fd, buf, BUFSIZE);
    write(0, buf, bytes_read);
    write(to_fd, buf, bytes_read);

    if(has_eof(buf, bytes_read))
      found_eof = 1;
    did_something = 1;
  }

  return did_something;
}


void transfer(int proxy_fd, int client_fd){
  int interval = 100;
  int max_idle = 50000;
  int cur_idle = 0;
  int found_eof = 0;

  while(1){
    if(pipe_async(proxy_fd, client_fd))
      cur_idle = 0;

    if(pipe_async(client_fd, proxy_fd))
      cur_idle = 0;

    usleep(interval);
    cur_idle += interval;

    if(found_eof || cur_idle > max_idle){
      if(found_eof)
	printf("Found eof\n");
      else if(cur_idle > max_idle)
	printf("cur_idle > max_idle\n");
      else
	printf("Don't know what I found\n");

      close(proxy_fd);
      close(client_fd);
      printf("session complete\n");
      fflush(stdout);
      exit(0);
    }
  }
}

int proxyechod(int client_fd){
  int proxy_fd = connectTCP(real_proxy, real_port);

  transfer(proxy_fd, client_fd);  

  return 0;
}

/*------------------------------------------------------------------------
 * reaper - clean up zombie children
 *------------------------------------------------------------------------
 */
/*ARGSUSED*/
void reaper(int sig){
  int	status;

  while (wait3(&status, WNOHANG, (struct rusage *)0) >= 0)
    /* empty */;
}
